﻿#include "rbtree.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <time.h>
#include <stdlib.h>


/*用于写日志*/
#define FILENAME "file_state.log"
FILE *plog; 
#define DEFINELOG /*注释掉就直接输出到标准出错中*/
#ifdef DEFINELOG 
#define OPENLOG(file_name) \
	if (NULL == (plog = fopen(file_name, "a"))) \
		plog = stderr; \
	setvbuf(plog, NULL, _IONBF, 0);
#define PRINTLOG(fmt, arg...) fprintf(plog, fmt, ##arg);fflush(plog)
#else
#define OPENLOG(file_name)
#define PRINTLOG(fmt, arg...) fprintf(stderr, fmt, ##arg)
#endif

/*用于调试*/
#define DEFINEDEBUG  /*注释掉就不能用了*/
#ifdef DEFINEDEBUG 
#define PRINTDEBUG(fmt, arg...) fprintf(stderr, fmt, ##arg)
#else
#define PRINTDEBUG(fmt, arg...) 
#endif


void rb_check(rb_node_t *rb_root)
{
	rb_node_t *rb_node, *pre_node, *temp_node;
	int black, count = 0, temp;


	for (rb_node = rb_first(rb_root); rb_node; rb_node = rb_next(rb_node))
	{
		if (!rb_node->lchild || !rb_node->rchild)	
		{
			temp = 0;
			temp_node = rb_node;
			while (temp_node)
			{
				if (BLACK == temp_node->color)
					temp ++;
				temp_node = temp_node->parent;
			}
			if (count && black != temp)
			{
				PRINTDEBUG("%s\n", "黑高度不等");	
				exit(-1);
			}
		}	
		if (count && key_compare(pre_node->key, rb_node->key) > 0)	
		{
			PRINTDEBUG("%s\n", "前面key大于后面");	
			exit(-1);
		}
		else
			black = temp;
		pre_node = rb_node;
		count ++;
	}
}
/*
 * ===  FUNCTION  ======================================================================
 *         Name:  get_file_name
 *  Description:  根据路径获取文件的名字
 * =====================================================================================
 */
char *get_file_name(char *path)
{
	int i;

	for (i = strlen(path) - 1; i >= 0 && path[i] != '/'; i --);
	return path + i + 1;
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  search
 *  Description:  搜索与key对应的节点
 * =====================================================================================
 */
rb_node_t *search(rb_key_t key, char *path, rb_node_t *root)
{
	rb_node_t *node;
	rb_node_t *llast; /*最后一个节点的下一个节点*/

	llast = rb_search_last(key, root);
	for (node = rb_search_first(key, root); node != llast; node = rb_next(node))
		if (0 == strcmp(node->info.path, path))
			return node;
	if (node && 0 == strcmp(node->info.path, path))
		return node;
	return NULL;
}
/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  scan_file
 *  Description:  扫描文件，将文件名作为关键字插入到红黑树中
 * =====================================================================================
 */
void scan_file(char *path, rb_node_t **root)
{
	struct stat stat_buf;	
	rb_node_t *temp;
	char *file_name;
	int len = strlen(path);
	if (len - 1 > 0 && path[len - 1] == '/')
		path[len - 1] = 0;

	/*判断文件的类型*/
	if (-1 == lstat(path, &stat_buf))
	{
		PRINTDEBUG("%s lstat error\n", path);	
		return ;
	}
	/*将当前文件插入到红黑树中*/
	temp = malloc(sizeof(rb_node_t));	
	file_name = get_file_name(path);	
	if ( 0 == strcmp(file_name, FILENAME)) /*当前文件为日志文件*/
		return ;
	/*检测当前文件是否已经加入红黑色*/
	if ((temp = search(file_name, path, *root)))
	{
		/*检测到了当前文件已经存在*/	
		/*判断是否已经修改*/
		if (temp->info.modify_time != stat_buf.st_mtime)
		{
			temp->info.modify_time = stat_buf.st_mtime;
			temp->info.flag = 2;
		}
		else
			temp->info.flag = 3;
	}
	else
	{
		temp = (rb_node_t *)malloc(sizeof(rb_node_t));
		strcpy(temp->key, file_name);
		strcpy(temp->info.path, path);
		temp->info.flag = 1;
		temp->info.modify_time = stat_buf.st_mtime;
		rb_insert_node(temp, root);
 		/*rb_check(*root);*/
	}	
	if (S_ISDIR(stat_buf.st_mode))
	{
		/*当前路径为目录*/	
		/*扫描目录中所有文件*/
		DIR *dir;
		struct dirent *dir_entry;
		
		if (NULL == (dir = opendir(path)))
		{
			PRINTDEBUG("%s opendir error\n", path);	
			return ;
		}
		len = strlen(path);
		path[len ++] = '/';
		path[len] = 0;
		while (NULL != (dir_entry = readdir(dir)))	
		{
			if (0 == strcmp(dir_entry->d_name, ".") || 0 == strcmp(dir_entry->d_name, ".."))
				continue;
			if (len + strlen(dir_entry->d_name) >= PATH_MAX)
			{
				PRINTDEBUG("path length is %d, big then PATHMAX\n", len);	
				return ;
			}
			strcpy(path + len, dir_entry->d_name);
			scan_file(path, root);
		}
		closedir(dir);
	}
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  check
 *  Description:  检测各个文件的标志，并将变化了的文件写入日志文件中
 *  			  flag:
 *  			  	1 : 新文件
 *  			  	2 : 文件被修改
 *  			  	3 : 文件没有被修改
 *  			  	4 : 文件被删除
 *  			  日志文件的格式：
 *  			  	文件路径 文件状态 状态修改的时间 
 * =====================================================================================
 */
void check(rb_node_t **root)
{
	rb_node_t *node, *next, *temp;
	char time_buf[21];
	time_t _time;
	static count = 1;

	node = rb_first(*root);
	while (node)
	{
		if (count == 1)
		{
			node->info.flag = 0;
			node = rb_next(node);
			continue;
		}
		if (3 != node->info.flag)
		{
			if (0 == node->info.flag)
				time(&_time);
			else
				_time = node->info.modify_time;
			strftime(time_buf, 21, "%F %T", localtime(&_time));
			PRINTLOG("%s\t", node->info.path);
			if (0 == node->info.flag)	
			{
				PRINTLOG("%s\t", "delete");
				/*从红黑树中删除该节点*/
				next = rb_next(node);
				temp = node;
				node = rb_erase_node(node, root);
				if (next == node)
				{
					node = temp;	
					free(next);
				}
				else
				{
					free(node);
					node = next;
				}
		/*		PRINTDEBUG("%s\n", "delete node");
 				rb_check(*root);*/
			}
			else
			{
				if (1 == node->info.flag)
				{
					PRINTLOG("%s\t", "new");
				}
				else 
				{
					PRINTLOG("%s\t", "modify");
				}
				node->info.flag = 0;
				node = rb_next(node);
			}
			PRINTLOG("%s\n", time_buf);
		}
		else
		{
			node->info.flag = 0;
			node = rb_next(node);
		}
	}
	count = 2;
}
int main ( int argc, char *argv[] ) 
{
	char path[PATH_MAX];
	/*打开日志*/
	OPENLOG(FILENAME)
	rb_node_t *root = NULL;
	if (argc != 2 || argv[1][0] != '/')
	{
		PRINTDEBUG("usage: %s path\n", argv[1]);	
		return -1;
	}
	while (1)
	{
		strcpy(path, argv[1]);
		scan_file(path, &root);	
		check(&root);
		sleep(1);
	}
	return 0;
}				/* ----------  end of function main  ---------- */
